

/* action.c */


/*------------------------------------------------------------------------
 * Module for the complete game calculation
 *------------------------------------------------------------------------
 * This version of action is part of the Solitaire Brute Force tool.
 * Copyright (C) 2001 Holger Hasselbach <H.Hasselbach@galad.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *------------------------------------------------------------------------
 */


/*----------------------------------------------------------------------*/

#include "action.h"
#include "mymemory.h"
#include "mystring.h"
#include "field.h"
#include "hashtabs.h"
#include "statgame.h"

#include <stdlib.h>
#include <stdio.h>

/*----------------------------------------------------------------------*/


struct s_ACTION
{
    tpsSTATGAME psStatGame;
    tpsHASHTABS psHashTabs;
    tpsFIELDVAL psFieldVal;
    tpsFIELDSTT psFieldStt;

    tpsFIELDVAL psNextVal;

    tPSZ        pzSpecials;
};

typedef struct s_ACTION tsACTION;

/*---------------------------------------------*/

static tVOID v_Action_PrintMemUsage (tpsACTION psAct);
static tVOID v_Action_PrintPattern  (tpsFIELDVAL psVal);
static tVOID v_Action_AddSpecials   (tpsACTION psAct, tpsFIELDVAL psVal);
static tVOID v_Action_CheckSpecials (tpsACTION psAct);


/*----------------------------------------------------------------------*/


/*-----------------------------------------------
 * Create a new ACTION
 *-----------------------------------------------
 */

tpsACTION psActionCreate (tVOID)
{
    tpsACTION psAct;

    psAct = pvMemoryAllocObject(sACTION);

    psAct->psStatGame = psStatGameCreate();
    psAct->psHashTabs = psHashTabsCreate();
    psAct->psFieldVal = psFieldValCreate();
    psAct->psFieldStt = psFieldSttCreate();

    psAct->psNextVal  = psFieldValCreate();

    psAct->pzSpecials = NULL;

    return (psAct);
}


/*-----------------------------------------------
 * Destroy an ACTION
 *-----------------------------------------------
 */

tpsACTION psActionDestroy (tpsACTION psAct)
{
    if (psAct)
      {
        psStatGameDestroy(psAct->psStatGame);
        psHashTabsDestroy(psAct->psHashTabs);
        psFieldValDestroy(psAct->psFieldVal);
        psFieldSttDestroy(psAct->psFieldStt);

        psFieldValDestroy(psAct->psNextVal);

        ppzStringClear(&(psAct->pzSpecials));

        pvMemoryFree(psAct);
      }

    return (NULL);
}


/*-----------------------------------------------
 * Memory statistics
 *-----------------------------------------------
 */

tpsACTION psActionMemUsage (tpsACTION psAct, tpCOUNT pcnSizeof, tpCOUNT pcnBlocks)
{
    if (psAct)
      {
        psStatGameMemUsage(psAct->psStatGame,pcnSizeof,pcnBlocks);
        psHashTabsMemUsage(psAct->psHashTabs,pcnSizeof,pcnBlocks);
        psFieldValMemUsage(psAct->psFieldVal,pcnSizeof,pcnBlocks);
        psFieldSttMemUsage(psAct->psFieldStt,pcnSizeof,pcnBlocks);

        psFieldValMemUsage(psAct->psNextVal ,pcnSizeof,pcnBlocks);

        ppzStringMemUsage (&(psAct->pzSpecials),pcnSizeof,pcnBlocks);

        *pcnSizeof += cnMemorySizeofObject(sACTION);
        *pcnBlocks += 1;
      }

    return (psAct);
}



/*---------------------------------------------------------
 * Game loop
 *---------------------------------------------------------
 */


/*-----------------------------------------------
 * Local: Print actual memory statistic
 *-----------------------------------------------
 */

static tVOID v_Action_PrintMemUsage (tpsACTION psAct)
{
    tCOUNT cnSizeof = 0,
           cnBlocks = 0;

    psActionMemUsage(psAct,&cnSizeof,&cnBlocks);
    vHashTabsMemEmpty(&cnSizeof,&cnBlocks);
    printf("Sze:%10lu Blk:%7lu ",(unsigned long)cnSizeof,(unsigned long)cnBlocks);
}


/*-----------------------------------------------
 * Local: Print field pattern
 *-----------------------------------------------
 */

static tVOID v_Action_PrintPattern (tpsFIELDVAL psVal)
{
    tPSZ pzPat = NULL;

    ppzFieldValPatternToStr(&pzPat,psVal);
    printf("%s",pzPat);
    ppzStringClear(&pzPat);
}


/*-----------------------------------------------
 * Local: Add pattern to Specials
 *-----------------------------------------------
 */

static tVOID v_Action_AddSpecials (tpsACTION psAct, tpsFIELDVAL psVal)
{
    tPSZ pzPat = NULL,
         pzStn = NULL;

    ppzFieldValPatternToStr(&pzPat,psVal);
    ppzStringULongToStr(&pzStn, (tULONG)cnFieldSttGetStones(psAct->psFieldStt) ,10);
    ppzStringConcatN(&(psAct->pzSpecials),4,psAct->pzSpecials,pzStn,":\n",pzPat);
    ppzStringClear(&pzPat);
    ppzStringClear(&pzStn);
}


/*-----------------------------------------------
 * Local: Check for specials
 *-----------------------------------------------
 */

static tVOID v_Action_CheckSpecials (tpsACTION psAct)
{
    if ( ( ( cnFieldSttGetMovesNrm(psAct->psFieldStt) == 0 ) &&
           ( cnFieldSttGetStones  (psAct->psFieldStt) > 20 )
         ) ||
           ( cnFieldSttGetSymType (psAct->psFieldStt) ==  1 ) ||
           ( cnFieldSttGetMovesNrm(psAct->psFieldStt) == 21 )
       )
        v_Action_AddSpecials(psAct,psAct->psFieldVal);
}


/*-----------------------------------------------
 * Calculation
 *-----------------------------------------------
 */

tVOID vActionGameLoop (tpsACTION psAct, tPSZc pzPattern)
{
    tCOUNT cnEntries,cnRound,cnNumber;

    vFieldValInitStart(psAct->psNextVal,pzPattern);
    printf("\n");
    v_Action_PrintPattern(psAct->psNextVal);
    v_Action_AddSpecials(psAct,psAct->psNextVal);

    if ( !bFieldValStartIsS1() )
        printf("Start is not symmetry s1, normalization disabled.\n");
    vHashTabsInsert(psAct->psHashTabs,psAct->psNextVal);
    cnEntries = 1;
    cnRound   = 0;

    vHashTabsOpen();

    printf("\n");
    while (cnEntries != 0)
      {
        cnRound++;
        v_Action_PrintMemUsage(psAct);
        printf("Rnd:%3lu Ent:%9lu ",(unsigned long)cnRound,(unsigned long)cnEntries);
        fflush(stdout);

        vHashTabsClose();
        printf("?"); fflush(stdout);
        vHashTabsOpen();

        cnNumber = 0;
        vHashTabsLoopInit(psAct->psHashTabs);
        while ( bHashTabsLoopNext(psAct->psHashTabs,psAct->psFieldVal) )
          {
            cnNumber++;
            if ((cnNumber & 0x3ffff) == 0)
              {
                printf("."); fflush(stdout);
              }

            vFieldSttCalculate(psAct->psFieldStt,psAct->psFieldVal);
            vStatGameAddField(psAct->psStatGame,psAct->psFieldStt);

            /* v_Action_CheckSpecials(psAct); */

            vFieldSttLoopInit(psAct->psFieldStt);
            while ( bFieldSttLoopNext(psAct->psFieldStt,psAct->psNextVal) )
                vHashTabsInsert(psAct->psHashTabs,psAct->psNextVal);
          }

        cnEntries = cnHashTabsEntriesList(psAct->psHashTabs);

        printf("!\n"); fflush(stdout);
      }

    vHashTabsClose();

    vStatGameCalcWhole(psAct->psStatGame);
}


/*-----------------------------------------------
 * Get specials
 *-----------------------------------------------
 */

tpPSZ ppzActionSpecials (tpPSZ ppzSpec, tpsACTION psAct)
{
    return( ppzStringCopy(ppzSpec,psAct->pzSpecials) );
}


/*----------------------------------------------------------------------*/


/*-----------------------------------------------
 * Create HTML table with statistic
 *-----------------------------------------------
 */

tpPSZ ppzActionTable (tpPSZ ppzTable, tpsACTION psAct, tPSZc pzContent)
{
    return ( ppzStatGameTable(ppzTable,psAct->psStatGame,pzContent) );
}


/*----------------------------------------------------------------------*/
