

/* field.h */


/*------------------------------------------------------------------------
 * Module for field arithmetics
 *------------------------------------------------------------------------
 */


/*
 * The structure of the field is completely hidden from the other modules;
 * any field arithmetics is calculated inside this module. To calculate any
 * other Solitaire-like games, only this module needs to be modified. Of
 * course, for other symmetry properties, a modification of the Stat modules
 * may be necessary, too.
 *
 * Three different structures are defined, due to the different access needs:
 * FIELDVAL holds a single field with the corresponding counters for the full
 *          and for the normalized tree. Thus FIELDVAL holds all informations
 *          necessary to describe a single field.
 * FIELDHSH holds an array with the same number of hash compressed fields
 *          as there are hash entries in a hash block, plus two arrays of
 *          counters for the full and the normalized counting. For any hash
 *          block, a FIELDHSH is allocated to store the FIELDVALs. Thus
 *          FIELDHSH holds all compressed informations necessary to save the
 *          fields in the hash table. It also holds any informations necessary
 *          to restore the FIELDVALs.
 *          To gather all important constants together in a single location,
 *          the hash constants are defined here, not in the hash module.
 * FIELDSTT holds the complete statistic informations about a single field
 *          and all the necessary informations about the following fields
 *          that can be reached by a single move. Thus FIELDSTT holds all the
 *          informations that can be calculated out of a FIELDVAL.
 * Therefore, the Hash module uses FIELDVAL and FIELDHSH for searching,
 * storing and restoring; the Stats module uses FIELDSTT for the complete
 * statistic; the Action module uses FIELDVAL to feed the Hash and FIELDSTT
 * to feed the Stats. Thus, all three structures are defined here without any
 * further modularization; any function has inner access to all three, no
 * matter to which it is applied.
 *
 * StonesRange  : There are 33 positions on the field, thus 0..33 stones.
 * SymIndexRange: There are 4 different symmetry types
 *                SymType :  1  2  4  8
 *                SymIndex:  0  1  2  3
 * NumberMoves  : There are 76 different possible moves.
 * MovesRange   : There can be 0..NumberMoves (in theory) possible moves
 *                on one single field.
 *
 * All range checkings are done with assert.
 */


/*----------------------------------------------------------------------*/

#ifndef FIELD_H
#define FIELD_H

/*----------------------------------------------------------------------*/

#include "mytypes.h"
#include "wordcnt.h"

/*----------------------------------------------------------------------*/


#define cnFieldHashRange     4096
#define cnFieldHashEntries    256

#define cnFieldStonesRange     34
#define cnFieldSymIndexRange    4
#define cnFieldNumberMoves     76
#define cnFieldMovesRange    ( cnFieldNumberMoves + 1 )


/*----------------------------------------------------------------------*/


typedef struct s_FIELDVAL *tpsFIELDVAL;
typedef tpsFIELDVAL       *tppsFIELDVAL;

/*---------------------------------------------*/

tpsFIELDVAL psFieldValCreate         (tVOID);
tpsFIELDVAL psFieldValDestroy        (tpsFIELDVAL psVal);
tpsFIELDVAL psFieldValMemUsage       (tpsFIELDVAL psVal, tpCOUNT pcnSizeof, tpCOUNT pcnBlocks);

tCOUNT      cnFieldValSymIndexToType (tCOUNT cnSymIndex);

tCOUNT      cnFieldValHashKey        (tpsFIELDVAL psVal);

tVOID       vFieldValInitStart       (tpsFIELDVAL psVal, tPSZc pzPattern);
tBOOL       bFieldValStartIsS1       (tVOID);
tBOOL       bFieldValIsWinning       (tpsFIELDVAL psVal);
tpPSZ       ppzFieldValPatternToStr  (tpPSZ ppzStr, tpsFIELDVAL psVal);


/*----------------------------------------------------------------------*/


typedef struct s_FIELDHSH *tpsFIELDHSH;
typedef tpsFIELDHSH       *tppsFIELDHSH;

/*---------------------------------------------*/

tpsFIELDHSH psFieldHshCreate   (tVOID);
tpsFIELDHSH psFieldHshDestroy  (tpsFIELDHSH psHsh);
tpsFIELDHSH psFieldHshMemUsage (tpsFIELDHSH psHsh, tpCOUNT pcnSizeof, tpCOUNT pcnBlocks);

tBOOL       bFieldHshSearch    (tpsFIELDHSH psHsh, tCOUNT cnEntries, tpsFIELDVAL psVal, tpCOUNT pcnIndex);
tVOID       vFieldHshSetField  (tpsFIELDHSH psHsh, tCOUNT cnIndex, tpsFIELDVAL psVal);
tVOID       vFieldHshGetField  (tpsFIELDHSH psHsh, tCOUNT cnIndex, tpsFIELDVAL psVal);
tVOID       vFieldHshAddField  (tpsFIELDHSH psHsh, tCOUNT cnIndex, tpsFIELDVAL psVal);

tVOID       vFieldHshSetKey    (tpsFIELDHSH psHsh, tCOUNT cnHashKey);
tCOUNT      cnFieldHshGetKey   (tpsFIELDHSH psHsh);


/*----------------------------------------------------------------------*/


typedef struct s_FIELDSTT *tpsFIELDSTT;
typedef tpsFIELDSTT       *tppsFIELDSTT;

/*---------------------------------------------*/

tpsFIELDSTT psFieldSttCreate      (tVOID);
tpsFIELDSTT psFieldSttDestroy     (tpsFIELDSTT psStt);
tpsFIELDSTT psFieldSttMemUsage    (tpsFIELDSTT psStt, tpCOUNT pcnSizeof, tpCOUNT pcnBlocks);

tVOID       vFieldSttCalculate    (tpsFIELDSTT psStt, tpsFIELDVAL psVal);

tVOID       vFieldSttLoopInit     (tpsFIELDSTT psStt);
tBOOL       bFieldSttLoopNext     (tpsFIELDSTT psStt, tpsFIELDVAL psVal);

tCOUNT      cnFieldSttGetMovesAll (tpsFIELDSTT psStt);
tCOUNT      cnFieldSttGetMovesNrm (tpsFIELDSTT psStt);
tCOUNT      cnFieldSttGetStones   (tpsFIELDSTT psStt);
tCOUNT      cnFieldSttGetSymType  (tpsFIELDSTT psStt);
tCOUNT      cnFieldSttGetSymIndex (tpsFIELDSTT psStt);
tBOOL       bFieldSttGetIsWinning (tpsFIELDSTT psStt);

tVOID       vFieldSttAddCountAll  (tpsFIELDSTT psStt, tpsWORDCNT psCount, tCOUNT cnIndex);
tVOID       vFieldSttAddCountNrm  (tpsFIELDSTT psStt, tpsWORDCNT psCount, tCOUNT cnIndex);


/*----------------------------------------------------------------------*/

#endif
