

/* hashblck.c */


/*------------------------------------------------------------------------
 * Module for management of a single hash block and a hash block list
 *------------------------------------------------------------------------
 * This version of hashblck is part of the Solitaire Brute Force tool.
 * Copyright (C) 2001 Holger Hasselbach <H.Hasselbach@galad.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *------------------------------------------------------------------------
 */


/*----------------------------------------------------------------------*/

#include "hashblck.h"
#include "mymemory.h"

/*----------------------------------------------------------------------*/


struct s_HASHBLCK
{
    tpsFIELDHSH psFields;
    tCOUNT      cnEntries;

    tpsHASHBLCK psNext;
};

typedef struct s_HASHBLCK tsHASHBLCK;


/*----------------------------------------------------------------------*/


/*-----------------------------------------------
 * Memory management variables
 *-----------------------------------------------
 */

static tpsHASHBLCK ps_HashBlck_EmptyList  = NULL;
static tBOOL       b_HashBlck_EmptyActive = FALSE;


/*-----------------------------------------------
 * Memory management open
 *-----------------------------------------------
 */

tVOID vHashBlckOpen (tVOID)
{
    if (!b_HashBlck_EmptyActive)
      {
        ps_HashBlck_EmptyList  = NULL;
        b_HashBlck_EmptyActive = TRUE;
      }
}


/*-----------------------------------------------
 * Memory management close
 *-----------------------------------------------
 */

tVOID vHashBlckClose (tVOID)
{
    if (b_HashBlck_EmptyActive)
      {
        b_HashBlck_EmptyActive = FALSE;
        ppsHashBlckListDestroy(&ps_HashBlck_EmptyList);
      }
}


/*-----------------------------------------------
 * Memory management statistics
 *-----------------------------------------------
 */

tVOID vHashBlckMemEmpty (tpCOUNT pcnSizeof, tpCOUNT pcnBlocks)
{
    if (b_HashBlck_EmptyActive)
        ppsHashBlckListMemUsage(&ps_HashBlck_EmptyList,pcnSizeof,pcnBlocks);
}


/*----------------------------------------------------------------------*/


/*-----------------------------------------------
 * Create a new HASHBLCK
 *-----------------------------------------------
 */

tpsHASHBLCK psHashBlckCreate (tCOUNT cnHashKey)
{
    tpsHASHBLCK psBlck;

    if ( b_HashBlck_EmptyActive && ps_HashBlck_EmptyList )
      {
        psBlck = ps_HashBlck_EmptyList;
        ps_HashBlck_EmptyList = psBlck->psNext;
      }
    else
      {
        psBlck = pvMemoryAllocObject(sHASHBLCK);
        psBlck->psFields = psFieldHshCreate();
      }

    vFieldHshSetKey(psBlck->psFields,cnHashKey);
    psBlck->cnEntries = 0;

    psBlck->psNext = NULL;

    return (psBlck);
}


/*-----------------------------------------------
 * Destroy a HASHBLCK
 *-----------------------------------------------
 */

tpsHASHBLCK psHashBlckDestroy (tpsHASHBLCK psBlck)
{
    tpsHASHBLCK psNext = NULL;

    if (psBlck)
      {
        psNext = psBlck->psNext;

        if (b_HashBlck_EmptyActive)
          {
            psBlck->psNext = ps_HashBlck_EmptyList;
            ps_HashBlck_EmptyList = psBlck;
          }
        else
          {
            psFieldHshDestroy(psBlck->psFields);
            pvMemoryFree(psBlck);
          }
      }

    return (psNext);
}


/*-----------------------------------------------
 * Memory statistics
 *-----------------------------------------------
 */

tpsHASHBLCK psHashBlckMemUsage (tpsHASHBLCK psBlck, tpCOUNT pcnSizeof, tpCOUNT pcnBlocks)
{
    if (psBlck)
      {
        psFieldHshMemUsage(psBlck->psFields,pcnSizeof,pcnBlocks);

        *pcnSizeof += cnMemorySizeofObject(sHASHBLCK);
        *pcnBlocks += 1;
      }

    return (psBlck);
}


/*----------------------------------------------------------------------*/


/*-----------------------------------------------
 * List: Destroy complete HASHBLCK-list
 *-----------------------------------------------
 */

tppsHASHBLCK ppsHashBlckListDestroy (tppsHASHBLCK ppsList)
{
    while (*ppsList)
        *ppsList = psHashBlckDestroy(*ppsList);

    return (ppsList);
}


/*-----------------------------------------------
 * List: Memory statistics
 *-----------------------------------------------
 */

tppsHASHBLCK ppsHashBlckListMemUsage (tppsHASHBLCK ppsList, tpCOUNT pcnSizeof, tpCOUNT pcnBlocks)
{
    tpsHASHBLCK psBlck;

    psBlck = *ppsList;
    while (psBlck)
      {
        psHashBlckMemUsage(psBlck,pcnSizeof,pcnBlocks);
        psBlck = psBlck->psNext;
      }

    return (ppsList);
}


/*-----------------------------------------------
 * List: Insert field
 *-----------------------------------------------
 * The field is searched in the hash block list.
 * If it exists, it is added to the existing entry.
 * If it not exists, it is newly inserted, with a
 * list extension if necessary.
 *-----------------------------------------------
 */

tVOID vHashBlckListInsert (tppsHASHBLCK ppsList, tpsFIELDVAL psVal)
{
    tBOOL       bFound = FALSE;
    tpsHASHBLCK psBlck = NULL;
    tCOUNT      cnIndex;

    while (!bFound && *ppsList)
      {
        psBlck = *ppsList;
        if ( bFieldHshSearch(psBlck->psFields,psBlck->cnEntries,psVal,&cnIndex) )
            bFound = TRUE;
        else
            ppsList = &((*ppsList)->psNext);
      }

    if (bFound)
        vFieldHshAddField(psBlck->psFields,cnIndex,psVal);
    else
      {
        if (!psBlck)
            psBlck = *ppsList = psHashBlckCreate( cnFieldValHashKey(psVal) );

        if (psBlck->cnEntries >= cnFieldHashEntries)
          {
            psBlck->psNext = psHashBlckCreate( cnFieldHshGetKey(psBlck->psFields) );
            psBlck = psBlck->psNext;
          }

        vFieldHshSetField(psBlck->psFields,psBlck->cnEntries,psVal);
        psBlck->cnEntries++;
      }
}


/*-----------------------------------------------
 * List: Concat a list to another list
 *-----------------------------------------------
 * All the Add-lists are added to a single List,
 * so List is added to the end of Add, to search
 * the end of Add, not the end of the longer and
 * longer List. Of course, the result is given in
 * List while Add is NULLed.
 *-----------------------------------------------
 */

tVOID vHashBlckListConcat (tppsHASHBLCK ppsList, tppsHASHBLCK ppsAdd)
{
    tppsHASHBLCK ppsBlck;

    ppsBlck = ppsAdd;
    while (*ppsBlck)
        ppsBlck = &((*ppsBlck)->psNext);

    *ppsBlck = *ppsList;
    *ppsList = *ppsAdd;
    *ppsAdd  = NULL;
}


/*-----------------------------------------------
 * List: Loop through the list, destroy empty blocks
 *-----------------------------------------------
 * Returns FALSE if the list is empty.
 *-----------------------------------------------
 */

tBOOL bHashBlckListLoop (tppsHASHBLCK ppsList, tpsFIELDVAL psVal)
{
    tBOOL bOk = FALSE;

    while ( *ppsList && ((*ppsList)->cnEntries == 0) )
        *ppsList = psHashBlckDestroy(*ppsList);

    if (*ppsList)
      {
        (*ppsList)->cnEntries--;
        vFieldHshGetField((*ppsList)->psFields,(*ppsList)->cnEntries,psVal);

        bOk = TRUE;
      }

    return (bOk);
}


/*-----------------------------------------------
 * List: Count total number of entries
 *-----------------------------------------------
 */

tCOUNT cnHashBlckListEntries (tppsHASHBLCK ppsList)
{
    tCOUNT cnEntries = 0;

    while (*ppsList)
      {
        cnEntries += (*ppsList)->cnEntries;
        ppsList = &((*ppsList)->psNext);
      }

    return (cnEntries);
}


/*----------------------------------------------------------------------*/
