

/* mytypes.h */


/*------------------------------------------------------------------------
 * my-library for the definition of the standard-types
 *------------------------------------------------------------------------
 * This version of mytypes is part of the Solitaire Brute Force tool.
 * Copyright (C) 2001 Holger Hasselbach <H.Hasselbach@galad.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *------------------------------------------------------------------------
 */


/*----------------------------------------------------------------------*/

#ifndef MYTYPES_H
#define MYTYPES_H

/*----------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>

/*----------------------------------------------------------------------*/


/*
 * Integer types
 *
 * The unsigned types must have at least the number of value bits as defined.
 * They may have more, but not less. This means for a tUBYTE ubB:
 *   ubB   = 255;
 *   ubB <<= 1;
 *   ubB >>= 1;
 * gives 127 on systems with a native 8 bit type (CHAR_BIT == 8), 255 on
 * systems without (CHAR_BIT > 8). In any assignment it must be assured that
 * any bit above the range of definition is set to 0.
 *   ubB = (ubB & 0x7f) << 1;
 *   ubB = (ubB << 1) & 0xff;
 */


/*--- BYTE - 8 Bit Integer - 0..0xff --------------------*/


typedef signed char      tSBYTE;                   /* sb */

typedef tSBYTE           *tpSBYTE;
typedef const tSBYTE     *tpcSBYTE;
typedef volatile tSBYTE  *tpvSBYTE;
typedef tpSBYTE          *tppSBYTE;


typedef unsigned char    tUBYTE;                   /* ub */

typedef tUBYTE           *tpUBYTE;
typedef const tUBYTE     *tpcUBYTE;
typedef volatile tUBYTE  *tpvUBYTE;
typedef tpUBYTE          *tppUBYTE;



/*--- WORD - 16 Bit Integer - 0..0xffff -----------------*/


typedef signed short int    tSWORD;                /* sw */

typedef tSWORD              *tpSWORD;
typedef const tSWORD        *tpcSWORD;
typedef volatile tSWORD     *tpvSWORD;
typedef tpSWORD             *tppSWORD;


typedef unsigned short int  tUWORD;                /* uw */

typedef tUWORD              *tpUWORD;
typedef const tUWORD        *tpcUWORD;
typedef volatile tUWORD     *tpvUWORD;
typedef tpUWORD             *tppUWORD;



/*--- LONG - 32 Bit Integer - 0..0xffffffff -------------*/


typedef signed long int    tSLONG;                 /* sl */

typedef tSLONG             *tpSLONG;
typedef const tSLONG       *tpcSLONG;
typedef volatile tSLONG    *tpvSLONG;
typedef tpSLONG            *tppSLONG;


typedef unsigned long int  tULONG;                 /* ul */

typedef tULONG             *tpULONG;
typedef const tULONG       *tpcULONG;
typedef volatile tULONG    *tpvULONG;
typedef tpULONG            *tppULONG;



/*--- INT - Undefined Common Integer --------------------*/


typedef signed int      tSINT;                     /* si */

typedef tSINT           *tpSINT;
typedef const tSINT     *tpcSINT;
typedef volatile tSINT  *tpvSINT;
typedef tpSINT          *tppSINT;


typedef unsigned int    tUINT;                     /* ui */

typedef tUINT           *tpUINT;
typedef const tUINT     *tpcUINT;
typedef volatile tUINT  *tpvUINT;
typedef tpUINT          *tppUINT;



/*
 * tCOUNT
 *
 * Properties:
 * - Unsigned integer
 * - At least 32 bit
 * - Compatible to full value range of size_t
 * - Compatible to full positive value range of ptrdiff_t
 *
 * Used for anything that has to do with memory and object counting:
 * - Sizeof for any type and array
 * - Size for malloc
 * - Indexing of arrays
 * - Number of objects
 *
 * Note: I'm aware that tCOUNT might not be the optimal type when speed is
 * a question. The int type is recommended to fit the CPUs properties best.
 * However, if the number of objects is given as a tCOUNT, the use of an
 * int as the loop variable in a for ( uiI = 0; uiI < cnNum; uiI++ ) sets
 * a limit to the maximum number of objects. Additional, it becomes an
 * infinity loop bug when cnNum becomes greater than the int range.
 * If it fits in memory, its size and its number fits in tCOUNT. Then, any
 * function should be able to handle it without artificial and arbitrary
 * limitations. Anything else is a question of system dependent micro-
 * optimizations.
 */

/*--- COUNT - Memory Counter ----------------------------*/


typedef tULONG           tCOUNT;                   /* cn */

typedef tCOUNT           *tpCOUNT;
typedef const tCOUNT     *tpcCOUNT;
typedef volatile tCOUNT  *tpvCOUNT;
typedef tpCOUNT          *tppCOUNT;



/*
 * VOID is not really a useful typedef, but included for consistence.
 */

/*--- VOID - Void Type ----------------------------------*/


typedef void            tVOID;                     /*  v */

typedef tVOID           *tpVOID;
typedef const tVOID     *tpcVOID;
typedef volatile tVOID  *tpvVOID;
typedef tpVOID          *tppVOID;



/*
 * BOOL
 *
 * Is 0 or 1. To a boolean variable, only the result of a logical expression
 * should be assigned, or the values FALSE (0) and TRUE (1). Because the
 * shortest available type (tUBYTE) is used for tBOOL, a direct assignment
 * can give false results by implicit casts.
 *   bFlag = ulI;                    (Fails for ulI==0xffffff00 and
 *                                    UBYTE = native 8 bit)
 *   bFlag = (ulI > 0);              (Works for any ulI)
 *   bFlag = ( ulI ? TRUE : FALSE ); (Explicit version)
 * It is commonly considered bad style and a source for bugs to directly test
 * against FALSE or TRUE.
 */

/*--- BOOL - Logical Boolean ----------------------------*/


typedef tUBYTE          tBOOL;                     /*  b */

typedef tBOOL           *tpBOOL;
typedef const tBOOL     *tpcBOOL;
typedef volatile tBOOL  *tpvBOOL;
typedef tpBOOL          *tppBOOL;


#define FALSE 0
#define TRUE  1



/*
 * CHAR and PSZ
 *
 * Encapsulation of the char type. It is compiler option dependent if it is
 * signed or unsigned. PSZ is a pointer to an array of CHAR and compatible
 * to the standard library string functions.
 */


/*--- CHAR - Basic Character ----------------------------*/


typedef char            tCHAR;                     /*  c */

typedef tCHAR           *tpCHAR;
typedef const tCHAR     *tpcCHAR;
typedef volatile tCHAR  *tpvCHAR;
typedef tpCHAR          *tppCHAR;



/*--- PSZ - Zero Terminated String ----------------------*/


typedef tpCHAR         tPSZ;                       /* pz */

typedef tPSZ           *tpPSZ;
typedef const tPSZ     *tpcPSZ;
typedef volatile tPSZ  *tpvPSZ;
typedef tpPSZ          *tppPSZ;

typedef tpcCHAR        tPSZc;
typedef tpvCHAR        tPSZv;



/*--- FILE - File stream structure ----------------------*/


typedef FILE            tFILE;                     /*  f */

typedef tFILE           *tpFILE;
typedef const tFILE     *tpcFILE;
typedef volatile tFILE  *tpvFILE;
typedef tpFILE          *tppFILE;



/*--- DOUBLE - Float double precision -------------------*/


typedef double            tDOUBLE;                 /* fd */

typedef tDOUBLE           *tpDOUBLE;
typedef const tDOUBLE     *tpcDOUBLE;
typedef volatile tDOUBLE  *tpvDOUBLE;
typedef tpDOUBLE          *tppDOUBLE;



/*----------------------------------------------------------------------*/

#endif
