

/* statgame.c */


/*------------------------------------------------------------------------
 * Solitaire field statistics: Game
 *------------------------------------------------------------------------
 * This version of statgame is part of the Solitaire Brute Force tool.
 * Copyright (C) 2001 Holger Hasselbach <H.Hasselbach@galad.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *------------------------------------------------------------------------
 */


/*----------------------------------------------------------------------*/

#include "statgame.h"
#include "mymemory.h"
#include "statstne.h"
#include "stattabs.h"
#include "wordcnt.h"
#include "mystring.h"

#include <assert.h>

/*----------------------------------------------------------------------*/


#define cn_StatGame_WinCnt 2
#define cn_StatGame_WinAll 0
#define cn_StatGame_WinNrm 1


struct s_STATGAME
{
    tpsSTATSTNE apsStones[cnFieldStonesRange];

    tpsSTATSTNE psStoneSum;

    tpsWORDCNT  psWaysWin;  /* (WinCnt) */
};

typedef struct s_STATGAME tsSTATGAME;

/*---------------------------------------------*/

static tVOID v_StatGame_TabSet      (tpsSTATGAME psGame, tpsSTATTABS psTabs, tPSZc pzContent);

static tVOID v_StatGame_TabCell     (tpsSTATGAME psGame, tpsSTATTABS psTabs, tPSZc pzContent, tPSZc pzGiven);
static tVOID v_StatGame_TabColStne  (tpsSTATGAME psGame, tpsSTATTABS psTabs, tPSZc pzContent);
static tVOID v_StatGame_TabColSymm  (tpsSTATGAME psGame, tpsSTATTABS psTabs, tPSZc pzContent);
static tVOID v_StatGame_TabColMove  (tpsSTATGAME psGame, tpsSTATTABS psTabs, tPSZc pzContent);
static tVOID v_StatGame_TabColumn   (tpsSTATGAME psGame, tpsSTATTABS psTabs, tPSZc pzContent);
static tVOID v_StatGame_TabLoopCols (tpsSTATGAME psGame, tpsSTATTABS psTabs, tPSZc pzContent);

static tVOID v_StatGame_TabRowStne  (tpsSTATGAME psGame, tpsSTATTABS psTabs, tPSZc pzContent);
static tVOID v_StatGame_TabRowSymm  (tpsSTATGAME psGame, tpsSTATTABS psTabs, tPSZc pzContent);
static tVOID v_StatGame_TabRowMove  (tpsSTATGAME psGame, tpsSTATTABS psTabs, tPSZc pzContent);
static tVOID v_StatGame_TabLoopRows (tpsSTATGAME psGame, tpsSTATTABS psTabs, tPSZc pzContent);


/*----------------------------------------------------------------------*/


/*-----------------------------------------------
 * Create a new STATGAME
 *-----------------------------------------------
 */

tpsSTATGAME psStatGameCreate (tVOID)
{
    tpsSTATGAME psGame;
    tCOUNT      cnI;

    psGame = pvMemoryAllocObject(sSTATGAME);

    for (cnI = 0; cnI < cnFieldStonesRange; cnI++)
        psGame->apsStones[cnI] = psStatStneCreate();

    psGame->psStoneSum = psStatStneCreate();

    psGame->psWaysWin = psWordCntCreate(cn_StatGame_WinCnt);
    vWordCntInit0All(psGame->psWaysWin);

    return (psGame);
}


/*-----------------------------------------------
 * Destroy a STATGAME
 *-----------------------------------------------
 */

tpsSTATGAME psStatGameDestroy (tpsSTATGAME psGame)
{
    tCOUNT cnI;

    if (psGame)
      {
        for (cnI = 0; cnI < cnFieldStonesRange; cnI++)
            psStatStneDestroy(psGame->apsStones[cnI]);

        psStatStneDestroy(psGame->psStoneSum);

        psWordCntDestroy(psGame->psWaysWin);

        pvMemoryFree(psGame);
      }

    return (NULL);
}


/*-----------------------------------------------
 * Memory statistics
 *-----------------------------------------------
 */

tpsSTATGAME psStatGameMemUsage (tpsSTATGAME psGame, tpCOUNT pcnSizeof, tpCOUNT pcnBlocks)
{
    tCOUNT cnI;

    if (psGame)
      {
        for (cnI = 0; cnI < cnFieldStonesRange; cnI++)
            psStatStneMemUsage(psGame->apsStones[cnI],pcnSizeof,pcnBlocks);

        psStatStneMemUsage(psGame->psStoneSum,pcnSizeof,pcnBlocks);

        psWordCntMemUsage(psGame->psWaysWin,pcnSizeof,pcnBlocks);

        *pcnSizeof += cnMemorySizeofObject(sSTATGAME);
        *pcnBlocks += 1;
      }

    return (psGame);
}


/*-----------------------------------------------
 * Add a field statistic
 *-----------------------------------------------
 */

tVOID vStatGameAddField (tpsSTATGAME psGame, tpsFIELDSTT psStt)
{
    tCOUNT cnStones;

    cnStones = cnFieldSttGetStones(psStt);
    assert(cnStones < cnFieldStonesRange);

    vStatStneAddField(psGame->apsStones[cnStones],psStt);

    if ( bFieldSttGetIsWinning(psStt) )
      {
        vFieldSttAddCountAll(psStt,psGame->psWaysWin,cn_StatGame_WinAll);
        vFieldSttAddCountNrm(psStt,psGame->psWaysWin,cn_StatGame_WinNrm);
      }
}


/*-----------------------------------------------
 * Add a STATGAME to another STATGAME
 *-----------------------------------------------
 * Unused but included for consistence.
 *-----------------------------------------------
 */

tVOID vStatGameAddObject (tpsSTATGAME psGame, tpsSTATGAME psAdd)
{
    tCOUNT cnI;

    for (cnI = 0; cnI < cnFieldStonesRange; cnI++)
        vStatStneAddObject(psGame->apsStones[cnI],psAdd->apsStones[cnI]);

    vStatStneAddObject(psGame->psStoneSum,psAdd->psStoneSum);

    vWordCntAdd(psGame->psWaysWin,cn_StatGame_WinAll,psAdd->psWaysWin,cn_StatGame_WinAll);
    vWordCntAdd(psGame->psWaysWin,cn_StatGame_WinNrm,psAdd->psWaysWin,cn_StatGame_WinNrm);
}


/*-----------------------------------------------
 * Calculate statistic
 *-----------------------------------------------
 */

tVOID vStatGameCalcWhole (tpsSTATGAME psGame)
{
    tCOUNT       cnI;
    tppsSTATSTNE ppsStone;

    ppsStone = psGame->apsStones;
    for (cnI = 0; cnI < cnFieldStonesRange; cnI++, ppsStone++)
      {
        vStatStneCalcWhole(*ppsStone);
        vStatStneAddObject(psGame->psStoneSum,*ppsStone);
      }
}


/*-----------------------------------------------
 * Test if there is any pattern at all
 *-----------------------------------------------
 * Unused but included for consistence.
 *-----------------------------------------------
 */

tBOOL bStatGameIsEmpty (tpsSTATGAME psGame)
{
    return ( bStatStneIsEmpty(psGame->psStoneSum) );
}


/*-----------------------------------------------
 * Get the maximum number of moves
 *-----------------------------------------------
 */

tCOUNT cnStatGameMaxMoves (tpsSTATGAME psGame)
{
    return ( cnStatStneMaxMoves(psGame->psStoneSum) );
}


/*----------------------------------------------------------------------*/


/*-----------------------------------------------
 * Local: Store parameters into table
 *-----------------------------------------------
 * NULL is not valid for pzContent
 * Content qualifier:
 * "S" : Stones[cnStone]
 * "Sn": Stones[n] with n = decimal number
 *       StoneSum if cnStone or n >= StonesRange
 * "+" : StoneSum
 * "Wa": WaysWin All
 * "Wn": WaysWin Nrm
 *-----------------------------------------------
 */

static tVOID v_StatGame_TabSet (tpsSTATGAME psGame, tpsSTATTABS psTabs, tPSZc pzContent)
{
    tCOUNT cnIndex;

    vStatTabsStyleEntry(psTabs);
    switch (*pzContent)
      {
      case 'S' :
        cnIndex   = cnStatTabsGetStone(psTabs);
        pzContent = pzStatTabsStrGetNumber(pzContent + 1,&cnIndex);
        if (cnIndex < cnFieldStonesRange)
          {
            vStatStneTabSet(psGame->apsStones[cnIndex],psTabs,pzContent);
          }
        else
          {
            vStatTabsStyleEntrySum(psTabs);
            vStatStneTabSet(psGame->psStoneSum,psTabs,pzContent);
          }
        break;

      case '+' :
        vStatTabsStyleEntrySum(psTabs);
        vStatStneTabSet(psGame->psStoneSum,psTabs,pzContent + 1);
        break;

      case 'W' :
        vStatTabsStyleEntryAdd(psTabs);
        vStatTabsStyleEntrySum(psTabs);
        switch (pzContent[1])
          {
          case 'a' :
            cnIndex = cn_StatGame_WinAll;
            vStatTabsStyleEntryAll(psTabs);
            break;
          case 'n' :
            cnIndex = cn_StatGame_WinNrm;
            vStatTabsStyleEntryNrm(psTabs);
            break;
          default : assert(FALSE);
          }
        vStatTabsSetCellWordCnt(psTabs,psGame->psWaysWin,cnIndex,pzContent[2]);
        break;

      default : assert(FALSE);
      }
}


/*-----------------------------------------------
 * Local: Set cell to text or stat entry
 *-----------------------------------------------
 * Text is identified by surrounding {}.
 * If Given is given, text is entered as StyleEnum,
 * else as StyleHeader.
 * NULL is not valid for pzContent.
 *-----------------------------------------------
 */

static tVOID v_StatGame_TabCell (tpsSTATGAME psGame, tpsSTATTABS psTabs, tPSZc pzContent, tPSZc pzGiven)
{
    tPSZ pzInner = NULL;

    if (*pzContent == '{')
      {
        pzStatTabsStrGetInner(pzContent,'{','}',&pzInner);
        if ( cnStringLength(pzGiven) > 0 )
          {
            vStatTabsStyleEnum(psTabs);
            ppzStringConcatN(&pzInner,3,pzGiven,( pzInner ? " " : "" ),pzInner);
          }
        else
            vStatTabsStyleHeader(psTabs);

        vStatTabsSetCellText(psTabs,pzInner);
        ppzStringClear(&pzInner);
      }
    else
        v_StatGame_TabSet(psGame,psTabs,pzContent);
}


/*-----------------------------------------------
 * Local: Column loop through Stones
 *-----------------------------------------------
 */

static tVOID v_StatGame_TabColStne (tpsSTATGAME psGame, tpsSTATTABS psTabs, tPSZc pzContent)
{
    tCOUNT cnI;
    tPSZ   pzStne = NULL;

    cnI = cnFieldStonesRange;
    while (cnI--)
        if ( !bStatStneIsEmpty(psGame->apsStones[cnI]) )
          {
            ppzStringULongToStr(&pzStne,(tULONG)cnI,10);

            vStatTabsSetStone(psTabs,cnI);
            v_StatGame_TabCell(psGame,psTabs,pzContent,pzStne);
          }

    ppzStringClear(&pzStne);
}


/*-----------------------------------------------
 * Local: Column loop through Symms
 *-----------------------------------------------
 */

static tVOID v_StatGame_TabColSymm (tpsSTATGAME psGame, tpsSTATTABS psTabs, tPSZc pzContent)
{
    tCOUNT cnI;
    tPSZ   pzSymm = NULL;

    for (cnI = 0; cnI < cnFieldSymIndexRange; cnI++)
      {
        ppzStringULongToStr(&pzSymm, (tULONG)cnFieldValSymIndexToType(cnI) ,10);
        ppzStringConcatN(&pzSymm,2,"s",pzSymm);

        vStatTabsSetSymm(psTabs,cnI);
        v_StatGame_TabCell(psGame,psTabs,pzContent,pzSymm);
      }

    ppzStringClear(&pzSymm);
}


/*-----------------------------------------------
 * Local: Column loop through Moves
 *-----------------------------------------------
 */

static tVOID v_StatGame_TabColMove (tpsSTATGAME psGame, tpsSTATTABS psTabs, tPSZc pzContent)
{
    tCOUNT cnI,cnMax;
    tPSZ   pzMove = NULL;

    cnMax = cnStatGameMaxMoves(psGame);
    for (cnI = 0; cnI <= cnMax; cnI++)
      {
        ppzStringULongToStr(&pzMove,(tULONG)cnI,10);

        vStatTabsSetMove(psTabs,cnI);
        v_StatGame_TabCell(psGame,psTabs,pzContent,pzMove);
      }

    ppzStringClear(&pzMove);
}


/*-----------------------------------------------
 * Local: Set one column entry
 *-----------------------------------------------
 * Entry can be a text or stat entry, optional
 * surrounded by a column loop:
 * Cs[Cell] : Loop through Stones
 * Cy[Cell] : Loop through Symms
 * Cm[Cell] : Loop through Moves
 * Without a loop, an empty pzGiven is given to
 * TabCell.
 * Cy[{Nrm}] defines 4 Enum cell entries:
 * "s1 Nrm", "s2 Nrm", "s4 Nrm", "s8 Nrm"
 * NULL is not valid for pzContent.
 *-----------------------------------------------
 */

static tVOID v_StatGame_TabColumn (tpsSTATGAME psGame, tpsSTATTABS psTabs, tPSZc pzContent)
{
    tPSZ pzInner = NULL;

    if (*pzContent == 'C')
      {
        assert( cnStringLength(pzContent) > 2 );
        pzStatTabsStrGetInner(pzContent + 2,'[',']',&pzInner);
        assert(pzInner != NULL);

        switch (pzContent[1])
          {
          case 's' : v_StatGame_TabColStne(psGame,psTabs,pzInner); break;
          case 'y' : v_StatGame_TabColSymm(psGame,psTabs,pzInner); break;
          case 'm' : v_StatGame_TabColMove(psGame,psTabs,pzInner); break;
          default  : assert(FALSE);
          }

        ppzStringClear(&pzInner);
      }
    else
        v_StatGame_TabCell(psGame,psTabs,pzContent,NULL);
}


/*-----------------------------------------------
 * Local: Loop through col entries
 *-----------------------------------------------
 * Single entries are separated by commas.
 *-----------------------------------------------
 */

static tVOID v_StatGame_TabLoopCols (tpsSTATGAME psGame, tpsSTATTABS psTabs, tPSZc pzContent)
{
    tPSZ pzPart = NULL;

    while (*pzContent)
      {
        pzContent = pzStatTabsStrGetSep(pzContent,',',&pzPart);
        assert(pzPart != NULL);

        v_StatGame_TabColumn(psGame,psTabs,pzPart);
      }

    ppzStringClear(&pzPart);
}


/*-----------------------------------------------
 * Local: Row loop through Stones
 *-----------------------------------------------
 */

static tVOID v_StatGame_TabRowStne (tpsSTATGAME psGame, tpsSTATTABS psTabs, tPSZc pzContent)
{
    tCOUNT cnI;

    cnI = cnFieldStonesRange;
    while (cnI--)
        if ( !bStatStneIsEmpty(psGame->apsStones[cnI]) )
          {
            vStatTabsStyleEnum(psTabs);
            vStatTabsSetCellULong(psTabs,(tULONG)cnI);

            vStatTabsSetStone(psTabs,cnI);
            v_StatGame_TabLoopCols(psGame,psTabs,pzContent);

            vStatTabsSetNextRow(psTabs);
          }

    vStatTabsStyleHeader(psTabs);
    vStatTabsSetCellText(psTabs,"Sum");

    vStatTabsSetStone(psTabs,cnFieldStonesRange);
    v_StatGame_TabLoopCols(psGame,psTabs,pzContent);

    vStatTabsSetNextRow(psTabs);
}


/*-----------------------------------------------
 * Local: Row loop through Symms
 *-----------------------------------------------
 */

static tVOID v_StatGame_TabRowSymm (tpsSTATGAME psGame, tpsSTATTABS psTabs, tPSZc pzContent)
{
    tCOUNT cnI;
    tPSZ   pzSymm = NULL;

    for (cnI = 0; cnI < cnFieldSymIndexRange; cnI++)
      {
        ppzStringULongToStr(&pzSymm, (tULONG)cnFieldValSymIndexToType(cnI) ,10);
        ppzStringConcatN(&pzSymm,2,"s",pzSymm);
        vStatTabsStyleEnum(psTabs);
        vStatTabsSetCellText(psTabs,pzSymm);

        vStatTabsSetSymm(psTabs,cnI);
        v_StatGame_TabLoopCols(psGame,psTabs,pzContent);

        vStatTabsSetNextRow(psTabs);
      }

    vStatTabsStyleHeader(psTabs);
    vStatTabsSetCellText(psTabs,"Sum");

    vStatTabsSetSymm(psTabs,cnFieldSymIndexRange);
    v_StatGame_TabLoopCols(psGame,psTabs,pzContent);

    vStatTabsSetNextRow(psTabs);

    ppzStringClear(&pzSymm);
}


/*-----------------------------------------------
 * Local: Column loop through Moves
 *-----------------------------------------------
 */

static tVOID v_StatGame_TabRowMove (tpsSTATGAME psGame, tpsSTATTABS psTabs, tPSZc pzContent)
{
    tCOUNT cnI,cnMax;

    cnMax = cnStatGameMaxMoves(psGame);
    for (cnI = 0; cnI <= cnMax; cnI++)
      {
        vStatTabsStyleEnum(psTabs);
        vStatTabsSetCellULong(psTabs,(tULONG)cnI);

        vStatTabsSetMove(psTabs,cnI);
        v_StatGame_TabLoopCols(psGame,psTabs,pzContent);

        vStatTabsSetNextRow(psTabs);
      }

    vStatTabsStyleHeader(psTabs);
    vStatTabsSetCellText(psTabs,"Sum");

    vStatTabsSetMove(psTabs,cnFieldMovesRange);
    v_StatGame_TabLoopCols(psGame,psTabs,pzContent);

    vStatTabsSetNextRow(psTabs);
}


/*-----------------------------------------------
 * Local: Loop through row entries
 *-----------------------------------------------
 * Rs(RowDef) : Loop through Stones
 * Ry(RowDef) : Loop through Symms
 * Rm(RowDef) : Loop through Moves
 * Rf(RowDef) : Free defined row
 * NULL is not valid for pzContent.
 *-----------------------------------------------
 */

static tVOID v_StatGame_TabLoopRows (tpsSTATGAME psGame, tpsSTATTABS psTabs, tPSZc pzContent)
{
    tPSZ  pzRowDef = NULL;
    tCHAR cDef;

    while (*pzContent)
      {
        assert( (*pzContent == 'R') && (cnStringLength(pzContent) > 2) );
        cDef = pzContent[1];
        pzContent = pzStatTabsStrGetInner(pzContent + 2,'(',')',&pzRowDef);
        assert(pzRowDef != NULL);

        switch (cDef)
          {
          case 's' : v_StatGame_TabRowStne(psGame,psTabs,pzRowDef); break;
          case 'y' : v_StatGame_TabRowSymm(psGame,psTabs,pzRowDef); break;
          case 'm' : v_StatGame_TabRowMove(psGame,psTabs,pzRowDef); break;

          case 'f' :
            v_StatGame_TabLoopCols(psGame,psTabs,pzRowDef);
            vStatTabsSetNextRow(psTabs);
            break;

          default  : assert(FALSE);
          }
      }

    ppzStringClear(&pzRowDef);
}


/*----------------------------------------------------------------------*/


/*-----------------------------------------------
 * Create HTML table with statistic
 *-----------------------------------------------
 * Content: "{Title}RowEntries"
 * Any {} define text. Inside of any text except
 * the Title, the following characters are not
 * allowed: , ( ) [ ] { }
 * For obvious reasons, inside the title the } is
 * not allowed.
 *-----------------------------------------------
 */

tpPSZ ppzStatGameTable (tpPSZ ppzTable, tpsSTATGAME psGame, tPSZc pzContent)
{
    tpsSTATTABS psTabs;
    tPSZ        pzTitle = NULL;

    ppzStringClear(ppzTable);

    if (pzContent)
      {
        psTabs = psStatTabsCreate();
        vStatTabsSetStart(psTabs);

        pzContent = pzStatTabsStrGetInner(pzContent,'{','}',&pzTitle);
        vStatTabsStyleTitle(psTabs);
        vStatTabsSetCellText(psTabs,pzTitle);
        vStatTabsSetNextRow(psTabs);
        ppzStringClear(&pzTitle);

        v_StatGame_TabLoopRows(psGame,psTabs,pzContent);

        ppzStatTabsHtmlStr(ppzTable,psTabs);

        psStatTabsDestroy(psTabs);
      }

    return (ppzTable);
}


/*----------------------------------------------------------------------*/
